@@ -1,8 +1,5 @@
 # -*- coding: utf-8 -*-
 # See LICENSE.txt for licensing terms
-#$URL: https://rst2pdf.googlecode.com/svn/tags/0.93/rst2pdf/basenodehandler.py $
-#$Date: 2012-02-28 21:07:21 -0300 (Tue, 28 Feb 2012) $
-#$Revision: 2443 $
 
 '''
 This module provides one useful class:  NodeHandler
@@ -34,47 +31,51 @@ type of docutils node, then default processing will occur and a warning
 will be logged.
 '''
 
-import types
+from copy import copy
 import inspect
-from log import log, nodeid
-from smartypants import smartyPants
+import types
+
+from smartypants import smartypants
 import docutils.nodes
-from flowables import BoundByWidth, TocEntry
+
+from .flowables import BoundByWidth, TocEntry
+from .log import log, nodeid
 
 
 class MetaHelper(type):
-    ''' MetaHelper is designed to generically enable a few of the benefits of
-        using metaclasses by encapsulating some of the complexity of setting
-        them up.
+    """MetaHelper is designed to generically enable a few of the benefits of
+    using metaclasses by encapsulating some of the complexity of setting
+    them up.
 
-        If a base class uses MetaHelper (by assigning __metaclass__ = MetaHelper),
-        then that class (and its metaclass inheriting subclasses) can control
-        class creation behavior by defining a couple of helper functions.
+    If a base class uses MetaHelper (by assigning __metaclass__ = MetaHelper),
+    then that class (and its metaclass inheriting subclasses) can control
+    class creation behavior by defining a couple of helper functions.
 
-        1) A base class can define a _classpreinit function.  This function
-           is called during __new__ processing of the class object itself,
-           but only during subclass creation (not when the class defining
-           the _classpreinit is itself created).
+    1) A base class can define a _classpreinit function.  This function
+       is called during __new__ processing of the class object itself,
+       but only during subclass creation (not when the class defining
+       the _classpreinit is itself created).
 
-           The subclass object does not yet exist at the time _classpreinit
-           is called.  _classpreinit accepts all the parameters of the
-           __new__ function for the class itself (not the same as the __new__
-           function for the instantiation of class objects!) and must return
-           a tuple of the same objects.  A typical use of this would be to
-           modify the class bases before class creation.
+       The subclass object does not yet exist at the time _classpreinit
+       is called.  _classpreinit accepts all the parameters of the
+       __new__ function for the class itself (not the same as the __new__
+       function for the instantiation of class objects!) and must return
+       a tuple of the same objects.  A typical use of this would be to
+       modify the class bases before class creation.
 
-        2) Either a base class or a subclass can define a _classinit() function.
-           This function will be called immediately after the actual class has
-           been created, and can do whatever setup is required for the class.
-           Note that every base class (but not every subclass) which uses
-           MetaHelper MUST define _classinit, even if that definition is None.
+    2) Either a base class or a subclass can define a _classinit() function.
+       This function will be called immediately after the actual class has
+       been created, and can do whatever setup is required for the class.
+       Note that every base class (but not every subclass) which uses
+       MetaHelper MUST define _classinit, even if that definition is None.
+
+     MetaHelper also places an attribute into each class created with it.
+     _baseclass is set to None if this class has no superclasses which
+     also use MetaHelper, or to the first such MetaHelper-using baseclass.
+     _baseclass can be explicitly set inside the class definition, in
+     which case MetaHelper will not override it.
+    """
 
-         MetaHelper also places an attribute into each class created with it.
-         _baseclass is set to None if this class has no superclasses which
-         also use MetaHelper, or to the first such MetaHelper-using baseclass.
-         _baseclass can be explicitly set inside the class definition, in
-         which case MetaHelper will not override it.
-    '''
     def __new__(clstype, name, bases, clsdict):
         # Our base class is the first base in the class definition which
         # uses MetaHelper, or None if no such base exists.
@@ -101,12 +102,13 @@ class MetaHelper(type):
             cls._classinit()
 
 
-class NodeHandler(object):
-    ''' NodeHandler classes are used to dispatch
-       to the correct class to handle some node class
-       type, via a dispatchdict in the main class.
-    '''
-    __metaclass__ = MetaHelper
+class NodeHandler(metaclass=MetaHelper):
+    """NodeHandler classes are used to dispatch
+    to the correct class to handle some node class
+    type, via a dispatchdict in the main class.
+    """
+
+    dispatchdict = {}
 
     @classmethod
     def _classpreinit(baseclass, clstype, name, bases, clsdict):
@@ -141,8 +143,10 @@ class NodeHandler(object):
                 t = repr(target)
                 old = repr(cls.dispatchdict[target])
                 new = repr(self)
-                log.debug('Dispatch handler %s for node type %s overridden by %s' %
-                    (old, t, new))
+                log.debug(
+                    'Dispatch handler %s for node type %s overridden by %s'
+                    % (old, t, new)
+                )
                 cls.dispatchdict[target] = self
 
     @staticmethod
@@ -156,11 +160,10 @@ class NodeHandler(object):
     def log_unknown(self, node, during):
         if not hasattr(self, 'unkn_node'):
             self.unkn_node = set()
-        cln=self.getclassname(node)
-        if not cln in self.unkn_node:
+        cln = self.getclassname(node)
+        if cln not in self.unkn_node:
             self.unkn_node.add(cln)
-            log.warning("Unkn. node (self.%s): %s [%s]",
-                during, cln, nodeid(node))
+            log.warning("Unkn. node (self.%s): %s [%s]", during, cln, nodeid(node))
             try:
                 log.debug(node)
             except (UnicodeDecodeError, UnicodeEncodeError):
@@ -173,7 +176,7 @@ class NodeHandler(object):
         try:
             log.debug("%s: %s", handlerinfo, node)
         except (UnicodeDecodeError, UnicodeEncodeError):
-            log.debug("%s: %r", handlerninfo, node)
+            log.debug("%s: %r", handlerinfo, node)
         log.debug("")
 
         # Dispatch to the first matching class in the MRO
@@ -189,7 +192,7 @@ class NodeHandler(object):
         return result
 
     def __call__(self, client):
-        ''' Get the dispatchers, wrapped up as methods for the client'''
+        '''Get the dispatchers, wrapped up as methods for the client'''
         textdispatch = types.MethodType(self.textdispatch, client)
         elemdispatch = types.MethodType(self.elemdispatch, client)
         return textdispatch, elemdispatch
@@ -204,16 +207,39 @@ class NodeHandler(object):
     def gather_elements(self, client, node, style):
         return client.gather_elements(node, style=style)
 
+    def _get_non_default_values_from_style(self, style):
+        """Return a dictionary of all the items in the style that are changed from their default value"""
+        items = style.__dict__.copy()
+        defaults = style.defaults
+        for key in defaults.keys():
+            if defaults[key] == items[key]:
+                del items[key]
+        return items
+
     def getstyle(self, client, node, style):
         try:
             if node['classes'] and node['classes'][0]:
-                # FIXME: Supports only one class, sorry ;-)
-                if client.styles.StyleSheet.has_key(node['classes'][0]):
-                    style = client.styles[node['classes'][0]]
-                else:
-                    log.info("Unknown class %s, ignoring. [%s]",
-                        node['classes'][0], nodeid(node))
-        except TypeError: # Happens when a docutils.node.Text reaches here
+                for n in range(len(node['classes'])):
+                    if node['classes'][n] in client.styles.StyleSheet:
+                        if n == 0:
+                            style = client.styles[node['classes'][n]]
+                        else:
+                            # merge the non-default properties of this style into the style we currently have
+                            style = copy(style)
+                            items = self._get_non_default_values_from_style(
+                                client.styles[node['classes'][n]]
+                            )
+                            items.pop("parent", None)
+                            name = f"{style.__dict__['name']}-{items['name']}"
+                            items['name'] = name
+                            style.__dict__.update(items)
+                    else:
+                        log.info(
+                            "Unknown class %s, ignoring. [%s]",
+                            n,
+                            nodeid(node),
+                        )
+        except TypeError:  # Happens when a docutils.node.Text reaches here
             pass
 
         if style is None or style == client.styles['bodytext']:
@@ -225,13 +251,12 @@ class NodeHandler(object):
         elements = self.gather_elements(client, node, style)
 
         # Make all the sidebar cruft unreachable
-        #if style.__dict__.get('float','None').lower() !='none':
-            #node.elements=[Sidebar(node.elements,style)]
-        #elif 'width' in style.__dict__:
+        # if style.__dict__.get('float','None').lower() !='none':
+        # node.elements=[Sidebar(node.elements,style)]
+        # elif 'width' in style.__dict__:
 
         if 'width' in style.__dict__:
-            elements = [BoundByWidth(style.width,
-                elements, style, mode="shrink")]
+            elements = [BoundByWidth(style.width, elements, style, mode="shrink")]
 
         return elements
 
@@ -244,13 +269,13 @@ class NodeHandler(object):
         try:
             for i in node['ids']:
                 client.pending_targets.append(i)
-        except TypeError: #Happens with docutils.node.Text
+        except TypeError:  # Happens with docutils.node.Text
             pass
 
         elements = self.getelements(client, node, style)
 
         if node.line and client.debugLinesPdf:
-            elements.insert(0,TocEntry(client.depth-1,'LINE-%s'%node.line))
+            elements.insert(0, TocEntry(client.depth - 1, 'LINE-%s' % node.line))
         node.elements = elements
         return elements
 
@@ -267,9 +292,12 @@ class NodeHandler(object):
 
     def apply_smartypants(self, text, smarty, node):
         # Try to be clever about when to use smartypants
-        if node.__class__ in (docutils.nodes.paragraph,
-                docutils.nodes.block_quote, docutils.nodes.title):
-            return smartyPants(text, smarty)
+        if node.__class__ in (
+            docutils.nodes.paragraph,
+            docutils.nodes.block_quote,
+            docutils.nodes.title,
+        ):
+            return smartypants(text, smarty)
         return text
 
     # End overridable attributes and methods for textdispatch
@@ -285,6 +313,6 @@ class NodeHandler(object):
         except UnicodeDecodeError:
             pass
 
-        text = self.apply_smartypants(text, client.smarty, node)
+        text = self.apply_smartypants(text, client.smartypants_attributes, node)
         node.pdftext = text
         return text
